/*
	Middle V1.1, extracts a part of a file
	Source code put in public domain by Didier Stevens, no Copyright
	http://didierstevens.com
	Use at your own risk

	Shortcommings, or todo's ;-)
	- no pipe support (file redirection)
	- file part must fit in memory
*/

#include <stdio.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <limits.h>

long ParseNumericArg(char *szArg)
{
	char *szError;
	long lResult;

	lResult = strtol(szArg, &szError, 0);
	if (*szError != '\0' || lResult == LONG_MIN || lResult == LONG_MAX)
		return -1;
	else
		return lResult;
}

main(int argc, char **argv)
{
	FILE *fIn, *fOut;
	struct stat statFile;
	long lStart, lLength;
	void *pBuffer;
	char *szFileIn;
	char *szStart;
	char *szLength;
	char *szFileOut;

	switch (argc)
	{
		case 4:
			szFileIn = argv[1];
			szStart = argv[2];
			szLength = NULL;
			szFileOut = argv[3];
			break;
		
		case 5:
			szFileIn = argv[1];
			szStart = argv[2];
			szLength = argv[3];
			szFileOut = argv[4];
			break;
		
		default:
			fprintf(stderr, "Usage: middle file-in start [length] file-out\n"
								 "Middle V1.1, extracts a part of a file\n"
								 "Copies 'length' number of bytes in 'file-in' starting at 'start' (0 indexed) to 'file-out'\n"
								 "'start' and 'length' can be decimal or hexadecimal (0x## notation)\n"
								 "'file-in' is copied until the last byte if 'length' is omitted\n"
								 "Source code put in the public domain by Didier Stevens, no Copyright\n"
								 "Use at your own risk\n"
								 "http://didierstevens.com\n");
			return -1;
	}

	if (stat(szFileIn, &statFile) != 0)
	{
		fprintf(stderr, "error opening file %s\n", szFileIn);
		return -1;
	}

	if ((lStart = ParseNumericArg(szStart)) == -1)
	{
		fprintf(stderr, "error in start argument %s\n", szStart);
		return -1;
	}

	if (lStart < 0)
	{
		fprintf (stderr, "start cannot be negative\n");
		return -1;
	}

	if (statFile.st_size <= lStart)
	{
		fprintf (stderr, "start %ld is larger than the size %ld of file %s\n", lStart, (long) statFile.st_size, szFileIn);
		return -1;
	}

	if (szLength == NULL)
	{
		lLength = statFile.st_size - lStart;
	}
	else
	{
		if ((lLength = ParseNumericArg(szLength)) == -1)
		{
			fprintf(stderr, "error in length argument %s\n", szLength);
			return -1;
		}

		if (lLength < 1)
		{
			fprintf (stderr, "length cannot be less than 1\n");
			return -1;
		}
	}

	if (statFile.st_size < lStart + lLength)
	{
		fprintf (stderr, "start + length %ld is larger than the size %ld of file %s\n", lStart + lLength, (long) statFile.st_size, szFileIn);
		return -1;
	}

	if ((fIn = fopen(szFileIn, "rb")) == NULL)
	{
		fprintf(stderr, "error opening file %s\n", szFileIn);
		return -1;
	}

	if (fseek(fIn, lStart, 0) != 0)
	{
		fprintf(stderr, "error fseek file %s %ld\n", szFileIn, lStart);
		fclose(fIn);
		return -1;
	}

	if ((pBuffer = malloc(lLength)) == NULL)
	{
		fprintf (stderr, "not enough memory\n");
		fclose(fIn);
		return -1;
	}

	if (fread(pBuffer, lLength, 1, fIn) != 1)
	{
		fprintf(stderr, "error reading file %s\n", szFileIn);
		fclose(fIn);
		free(pBuffer);
		return -1;
	}

	fclose(fIn);

	if ((fOut = fopen(szFileOut, "wb")) == NULL)
	{
		fprintf(stderr, "error opening file %s\n", szFileOut);
		free(pBuffer);
		return -1;
	}

	if (fwrite(pBuffer, lLength, 1, fOut) != 1)
	{
		fprintf(stderr, "error writing file %s\n", szFileOut);
		fclose (fOut);
		free(pBuffer);
		return -1;
	}

	fclose (fOut);
	free(pBuffer);

	return 0;
}
